class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

    @classmethod
    def delete_instances_for_type(mcs, type_to_delete: type):
        if type_to_delete in mcs._instances:
            inst = mcs._instances.pop(type_to_delete)  # noqa
            del inst



from typing import List

from octave_sdk.grpc.quantummachines.octave.api.v1 import (
    SynthUpdate,
    SingleUpdate,
    RfUpConvUpdate,
    RfDownConvUpdate,
    IfDownConvUpdate,
    ClockUpdate,
)


class _BuilderList:
    def __init__(self) -> None:
        super().__init__()
        self._items = {}

    def get_items(self) -> List:
        return list(self._items.values())


class _SynthBuilderList(_BuilderList):
    def __init__(self) -> None:
        super().__init__()

    def __getitem__(self, item) -> SynthUpdate:
        if item not in self._items:
            self._items[item] = SingleUpdate(synth=SynthUpdate(index=item))
        return self._items[item].synth


class _DownBuilderList(_BuilderList):
    def __init__(self) -> None:
        super().__init__()

    def __getitem__(self, item) -> RfDownConvUpdate:
        if item not in self._items:
            self._items[item] = SingleUpdate(rf_down_conv=RfDownConvUpdate(index=item))
        return self._items[item].rf_down_conv


class _UpBuilderList(_BuilderList):
    def __init__(self) -> None:
        super().__init__()

    def __getitem__(self, item) -> RfUpConvUpdate:
        if item not in self._items:
            self._items[item] = SingleUpdate(rf_up_conv=RfUpConvUpdate(index=item))
        return self._items[item].rf_up_conv


class _IFBuilderList(_BuilderList):
    def __init__(self) -> None:
        super().__init__()

    def __getitem__(self, item) -> IfDownConvUpdate:
        if item not in self._items:
            self._items[item] = SingleUpdate(if_down_conv=IfDownConvUpdate(index=item))
        return self._items[item].if_down_conv


class _ClockBuilderList(_BuilderList):
    def __init__(self) -> None:
        super().__init__()

    def __getitem__(self, item) -> ClockUpdate:
        if item not in self._items:
            self._items[item] = SingleUpdate(clock=ClockUpdate())
        return self._items[item].clock_dist


class ClientRequestBuilder:
    def __init__(self) -> None:
        super().__init__()
        self._synth_builder_list = _SynthBuilderList()
        self._up_builder_list = _UpBuilderList()
        self._down_builder_list = _DownBuilderList()
        self._if_builder_list = _IFBuilderList()
        self._clock_builder_list = _ClockBuilderList()
        self._updates = []

    def __getattr__(self, item):
        if item == "synth":
            return self._synth_builder_list
        if item == "up":
            return self._up_builder_list
        if item == "down":
            return self._down_builder_list
        if item == "ifconv":
            return self._if_builder_list
        if item == "clk":
            return self._clock_builder_list

    def get_updates(self):
        return (
            self._synth_builder_list.get_items()
            + self._down_builder_list.get_items()
            + self._up_builder_list.get_items()
            + self._if_builder_list.get_items()
            + self._clock_builder_list.get_items()
        )



import asyncio
import copy
import dataclasses
import logging
import time
from pprint import PrettyPrinter
from typing import Optional, List, Dict

from grpclib.client import Channel, Configuration
from grpclib.events import SendRequest, listen
import betterproto
from octave_sdk.octave_async_thread import EventLoopThread, ConnectionDetails

from octave_sdk.batch import Batched, BatchSingleton
from octave_sdk.grpc.quantummachines.octave.api.v1 import (
    OctaveServiceStub,
    OctaveModule,
    ModuleReference,
    ControlRequest,
    UpdateRequest,
    SaveRequest,
    AquireRequest,
    RecallRequest,
    ListRequest,
    GetVersionRequest,
    MonitorRequest,
    ExploreRequest,
    IdentifyRequest,
    MonitorResponse,
    SingleUpdate,
    MonitorResponseModuleStatusError,
    ResetRequest,
    AquireResponse,
    ControlResponseRdataDebug,
)

logger = logging.getLogger("qm")

"""
# OctaveClient
#
#   This is a gRPC based Octave. The standard setup is the whole Octave product.
#   Namely, we access the SOM on its motherboard via ethernet and from there
#   communicate with all the boards via the main PIC.
#
#   A temporary alternative is using a mini SOM eval board which is connected
#   to a mini motherboard, directly into the serial lines of its PIC.
#
#   The specific variation is passed to the constructor.
"""


module_type_to_class = {
    OctaveModule.OCTAVE_MODULE_RF_UPCONVERTER: "rf_up_conv",
    OctaveModule.OCTAVE_MODULE_RF_DOWNCONVERTER: "rf_down_conv",
    OctaveModule.OCTAVE_MODULE_SYNTHESIZER: "synth",
    OctaveModule.OCTAVE_MODULE_IF_DOWNCONVERTER: "if_down_conv",
    OctaveModule.OCTAVE_MODULE_MOTHERBOARD: "motherboard",
}


class DebugSetException(Exception):
    pass


def _build_grpclib_channel(host, port, octave_name: Optional = None, credentials: Optional = None) -> Channel:
    channel = Channel(
        host=host,
        port=port,
        ssl=credentials is not None,
        config=Configuration(),
    )

    if octave_name is not None:

        async def send_request(event: SendRequest):
            event.metadata["x-grpc-service"] = octave_name

        listen(channel, SendRequest, send_request)

    return channel


@dataclasses.dataclass
class MonitorData:
    temp: float
    errors: List[MonitorResponseModuleStatusError]


class MonitorResult:
    def __init__(self, results_pb: MonitorResponse):
        self.modules: Dict[OctaveModule, List[MonitorData]] = {}
        self.modules = {
            OctaveModule.OCTAVE_MODULE_RF_UPCONVERTER: [None for _ in range(5)],
            OctaveModule.OCTAVE_MODULE_RF_DOWNCONVERTER: [None for _ in range(2)],
            OctaveModule.OCTAVE_MODULE_IF_DOWNCONVERTER: [None for _ in range(2)],
            OctaveModule.OCTAVE_MODULE_SYNTHESIZER: [None for _ in range(6)],
            OctaveModule.OCTAVE_MODULE_SOM: [None],
        }
        for module in results_pb.modules:
            self.modules[module.module.type][module.module.index - 1] = MonitorData(
                temp=module.temperature, errors=module.errors
            )

    def __repr__(self):
        pp = PrettyPrinter(width=150)
        return pp.pformat(self.modules)

    def __eq__(self, other):
        # Basically compare all elements except temp
        self_copy = copy.deepcopy(self)
        self._clear_temp_values(self_copy)
        other_copy = copy.deepcopy(other)
        self._clear_temp_values(other_copy)

        return self_copy.modules == other_copy.modules

    def _clear_temp_values(self, monitor_result):
        for module_list in monitor_result.modules.values():
            for monitor_data in module_list:
                if monitor_data is not None:
                    monitor_data.temp = 0


class ExploreResult:
    def __init__(self, results_pb):
        self.modules = {
            OctaveModule.OCTAVE_MODULE_RF_UPCONVERTER: [None for _ in range(5)],
            OctaveModule.OCTAVE_MODULE_RF_DOWNCONVERTER: [None for _ in range(2)],
            OctaveModule.OCTAVE_MODULE_IF_DOWNCONVERTER: [None for _ in range(2)],
            OctaveModule.OCTAVE_MODULE_SYNTHESIZER: [None for _ in range(6)],
            OctaveModule.OCTAVE_MODULE_MOTHERBOARD: [None for _ in range(1)],
        }
        for module in results_pb.modules:
            self.modules[module.module.type][module.module.index - 1] = module.id

    def __repr__(self):
        module_types = [
            OctaveModule.OCTAVE_MODULE_RF_UPCONVERTER,
            OctaveModule.OCTAVE_MODULE_RF_DOWNCONVERTER,
            OctaveModule.OCTAVE_MODULE_IF_DOWNCONVERTER,
            OctaveModule.OCTAVE_MODULE_SYNTHESIZER,
        ]

        m_id = self.modules[OctaveModule.OCTAVE_MODULE_MOTHERBOARD][0]
        if m_id is None or m_id == "":
            res = "MOTHERBOARD [\x1b[38;5;208m???\x1b[0m]\n"
        else:
            res = f"MOTHERBOARD [\x1b[38;5;154m{m_id}\x1b[0m]\n"

        res += "    RF_UPCONVs          RF_DOWNCONVs        IF_DOWNCONVs        " "SYNTHESIZERs\n"
        for index in range(6):
            res += "     "
            for t in module_types:
                if index < len(self.modules[t]):
                    m_id = self.modules[t][index]
                    if m_id is None:
                        res += f"\x1b[38;5;241m{index + 1}. {'---':17s}\x1b[0m"
                    elif m_id == "":
                        res += f"{index + 1}. \x1b[38;5;208m{'???':17s}\x1b[0m"
                    else:
                        res += f"{index + 1}. \x1b[38;5;154m{m_id:17s}\x1b[0m"
                else:
                    res += f"{'':20s}"
            res += "\n"
        return res


async def create_channel(details):
    channel = Channel(
        host=details.host,
        port=details.port,
        ssl=details.credentials is not None,
        config=Configuration(
            http2_connection_window_size=details.max_message_size,
            http2_stream_window_size=details.max_message_size,
        ),
    )
    return channel


class OctaveClient(Batched):
    def __init__(self, host: str, port: int, octave_name: str = None):
        self._host = host
        self._port = port
        self._octave_name = octave_name
        self._event_loop = EventLoopThread().loop
        super().__init__()

        connection_details = ConnectionDetails(host, port, credentials=None)
        self._timeout = connection_details.timeout

        self._channel = asyncio.run_coroutine_threadsafe(
            create_channel(connection_details), loop=self._event_loop
        ).result()
        self._service = OctaveServiceStub(self._channel)

    def __hash__(self):
        return hash((self._octave_name, self._host, self._port))

    @property
    def name(self) -> str:
        return self._octave_name

    def __del__(self):
        self._service = None
        self._channel.close()

    def _control(self, w_data=b"", r_length=0):
        control_request = ControlRequest(w_data=w_data, r_length=r_length)
        future = asyncio.run_coroutine_threadsafe(self._service.control(control_request), loop=self._event_loop)
        return future.result()

    def _format_cached_modules(self, modules: AquireResponse):
        cached_modules = {}
        for update in modules.state.updates:
            module_type, message = betterproto.which_one_of(update, "update")

            if module_type not in ("clock", "motherboard"):
                update_id = message.index
            else:
                update_id = 0

            cache_key = (update_id, module_type)
            cached_modules[cache_key] = update
        BatchSingleton().set_cached_modules(self, cached_modules)

    def _start_batch_callback(self):
        self._format_cached_modules(self.aquire_modules([]))

    def _end_batch_callback(self):
        self._send_update(list(BatchSingleton().get_cached_updates(self).values()))

    def update(self, updates: List[SingleUpdate]):
        if BatchSingleton().is_batch_mode:
            for update in updates:
                self._cache_update(update)
        else:
            self._send_update(updates)

    def _cache_update(self, update: SingleUpdate):
        module_type, message = betterproto.which_one_of(update, "update")

        if module_type not in ("clock", "motherboard"):
            update_id = message.index
        else:
            update_id = 0

        cache_key = (update_id, module_type)
        current_updates = BatchSingleton().get_cached_updates(self)
        if current_updates.get(cache_key):
            previous_cached_update = getattr(current_updates.get(cache_key), module_type).to_dict()
        else:
            previous_cached_update = {}

        previous_cached_update.update(message.to_dict())
        message.from_dict(previous_cached_update)

        new_update = SingleUpdate()
        setattr(new_update, module_type, message)
        current_updates[cache_key] = new_update
        BatchSingleton().set_cached_updates(self, current_updates)

    def _send_update(self, updates: List[SingleUpdate]):
        update_request = UpdateRequest(updates=updates)
        future = asyncio.run_coroutine_threadsafe(self._service.update(update_request), loop=self._event_loop)
        response = future.result()
        if not response.success:
            raise Exception(f"Octave update failed: {response.error_message}")
        return response

    def debug_request_clock_print(self):
        self._control(
            w_data=bytes([0xFF, 0xB9]),
            r_length=1,
        )

    def debug_set(
        self,
        monitor_enabled: bool = None,
        monitor_timeout: int = None,
        monitor_print_rate: int = None,
        monitor_update_fan: bool = None,
        uart_debug_mode: bool = None,
        print_updates: bool = None,
        min_fan_speed: bool = None,
        min_temp: int = None,
        max_temp_modules: int = None,
        max_temp_fpga: int = None,
    ):
        if monitor_timeout is not None:
            if monitor_timeout < 1 or monitor_timeout > 15:
                print("OctaveClientBase.debug_set   ERROR    monitor_timeout should be" " 1..15")
                return
        else:
            monitor_timeout = 0x00

        if monitor_print_rate is not None:
            if monitor_print_rate < 0 or monitor_print_rate > 255:
                print(
                    "OctaveClientBase.debug_set   ERROR    monitor_print_rate should either 0 (no printings) or 1..255"
                )
                return

        activate = 0x00
        state = 0x00

        if monitor_enabled is not None:
            activate |= 0x01
            state |= 0x01 if monitor_enabled else 0x00

        if uart_debug_mode is not None:
            activate |= 0x02
            state |= 0x02 if uart_debug_mode else 0x00

        if print_updates is not None:
            activate |= 0x04
            state |= 0x04 if print_updates else 0x00

        if monitor_print_rate is not None:
            activate |= 0x08
        else:
            monitor_print_rate = 0

        if monitor_update_fan is not None:
            activate |= 0x10
            state |= 0x10 if monitor_update_fan else 0x00

        if min_fan_speed is not None:
            activate |= 0x20
            min_fan_speed = min(int(min_fan_speed), 31)
            state |= (min_fan_speed & 1) << 5
            monitor_timeout |= (min_fan_speed & 0x1E) << 3

        if min_temp is not None:
            if min_temp < 1 or min_temp > 60:
                print("OctaveClientBase.debug_set   ERROR    min_temp should be between 1 to 60")
                return
            activate |= 0x40
        else:
            min_temp = 0x00

        if (max_temp_modules and not max_temp_fpga) or (max_temp_fpga and not max_temp_modules):
            logger.error("max_temp_modules and max_temp_fpga must come in paris")
            return

        if max_temp_modules and max_temp_fpga:
            if max_temp_modules < 1 or max_temp_modules > 65:
                print("OctaveClientBase.debug_set   ERROR    max_temp_modlues should be between 1 to 65")
                return
            if max_temp_fpga < 1 or max_temp_fpga > 75:
                print("OctaveClientBase.debug_set   ERROR    max_temp_fpga should be between 1 to 75")
                return
            activate |= 0x80
        else:
            max_temp_modules = 0
            max_temp_fpga = 0

        control_payload_old: bytes = bytes(
            [
                0xFF,
                0xFF,
                activate,
                state,
                monitor_timeout,
                monitor_print_rate,
            ]
        )

        control_payload_new = control_payload_old + bytes(
            [
                min_temp,
                max_temp_modules,
                max_temp_fpga,
            ]
        )

        # Try new format
        res = (
            self._control(
                w_data=control_payload_new,
                r_length=1,
            )
        ).r_data
        res_int = int.from_bytes(res, "little")
        if res_int == ControlResponseRdataDebug.RDATA_DEBUG_SUCCESS_RESPONSE:
            return

        # Fall back to old format
        res = (
            self._control(
                w_data=control_payload_old,
                r_length=1,
            )
        ).r_data
        res_int = int.from_bytes(res, "little")
        if res_int != ControlResponseRdataDebug.RDATA_DEBUG_SUCCESS_RESPONSE:
            raise DebugSetException("Failed to set debug params")

    def save_modules(self, m_id: str = None, module_refs=None, overwrite=True):
        module_id = m_id or "default"
        module_references = module_refs if module_refs else []
        save_request = SaveRequest(
            id=module_id, modules=module_references, overwrite=overwrite, timestamp=int(time.time())
        )

        future = asyncio.run_coroutine_threadsafe(self._service.save(save_request), loop=self._event_loop)
        return future.result()

    def _fetch_aquire_from_cache(self, modules: List[ModuleReference]):
        res = AquireResponse(state=UpdateRequest(updates=[]))
        for module in modules:
            module_type = module_type_to_class[module.type]
            if module_type != "motherboard":
                update_id = module.index
            else:
                update_id = 0

            cache_key = (update_id, module_type)
            cached_modules = BatchSingleton().get_cached_modules(self)
            cached_updates = BatchSingleton().get_cached_updates(self)
            if cache_key in cached_modules:
                cached_module = cached_modules[cache_key].to_dict()
                cached_update = cached_updates.get(cache_key, SingleUpdate()).to_dict()

                cached_module.update(cached_update)
                new_update = SingleUpdate().from_dict(cached_module)
                res.state.updates.append(new_update)
        return res

    def aquire_modules(self, modules: List[ModuleReference], use_cache=True):
        if BatchSingleton().is_batch_mode:
            res = self._fetch_aquire_from_cache(modules)

        else:
            aquire_request = AquireRequest(modules=modules, use_cache=use_cache)
            future = asyncio.run_coroutine_threadsafe(self._service.aquire(aquire_request), loop=self._event_loop)
            res = future.result()

        if len(modules) == 1 and len(res.state.updates) == 1:
            return res.state.updates[0]

        return res

    def recall(self, m_id: str = None):
        recall_request = RecallRequest(id="default" if m_id is None else m_id)
        future = asyncio.run_coroutine_threadsafe(self._service.recall(recall_request), loop=self._event_loop)
        return future.result()

    def configs(self):
        future = asyncio.run_coroutine_threadsafe(self._service.list(ListRequest()), loop=self._event_loop)
        return future.result().save_infos

    def version(self):
        future = asyncio.run_coroutine_threadsafe(self._service.get_version(GetVersionRequest()), loop=self._event_loop)
        return future.result()

    def monitor(self, sense_only=True) -> MonitorResult:
        monitor_request = MonitorRequest(sense_only=sense_only)

        future = asyncio.run_coroutine_threadsafe(self._service.monitor(monitor_request), loop=self._event_loop)
        return MonitorResult(future.result())

    def explore(self):
        future = asyncio.run_coroutine_threadsafe(self._service.explore(ExploreRequest()), loop=self._event_loop)
        return ExploreResult(future.result())

    def identify(self):
        future = asyncio.run_coroutine_threadsafe(self._service.identify(IdentifyRequest()), loop=self._event_loop)
        return future.result()

    def reset(self) -> bool:
        future = asyncio.run_coroutine_threadsafe(self._service.reset(ResetRequest()), loop=self._event_loop)
        return future.result()

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def octave_name(self):
        return self._octave_name



import betterproto
from enum import Enum
from octave_sdk.grpc.quantummachines.octave.api.v1 import (
    RfSource,
    IfSource,
    SynthRfOutput,
    RfUpConvIdentity,
    RfDownConvIdentity,
    IfDownConvIdentity,
    SynthIdentity,
    OctaveModule,
)

from octave_sdk.connectivity.exceptions import (
    ConnectivityException,
    InvalidIdentityException,
    RfSourceTypeException,
    ExternalInputTypeException,
)
from octave_sdk.connectivity.connectivity_util import RFInputRFSource


# Found in grpc, but not in betterproto :/, should be used for connectivity identify parsing and NOT grpc communication
def clear_rfsource_synth_output_oneof(rf_source: RfSource):
    rf_source.synth_output.index = 0
    rf_source.synth_output.output_port = 0


# Found in grpc, but not in betterproto :/, should be used for connectivity identify parsing and NOT grpc communication
def clear_rfsource_up_conv_oneof(rf_source: RfSource):
    rf_source.rf_up_conv_output.index = 0


# Found in grpc, but not in betterproto :/, should be used for connectivity identify parsing and NOT grpc communication
def clear_ifsource_down_conv_oneof(if_source: IfSource):
    if_source.rf_downconv_if_source.index = 0
    if_source.if_downconv_if_source.channel_index = 0


# Convert gRPC RfSource oneof to local enum
class RfSourceType(str, Enum):
    rf_up_conv_output = "rf_up_conv_output"
    synth_output = "synth_output"
    external_input = "external_input"
    constant_source = "constant_source"


# Convert gRPC IfSource oneof to local enum
class IfSourceType(str, Enum):
    rf_downconv_if_source = "rf_downconv_if_source"
    if_downconv_if_source = "if_downconv_if_source"
    external_if_input = "external_if_input"
    constant_source = "constant_source"


def get_rf_source_type(rf_source: RfSource) -> RfSourceType:
    one_of, _ = betterproto.which_one_of(rf_source, "source")
    return one_of


def get_if_source_type(if_source: IfSource) -> IfSourceType:
    one_of, _ = betterproto.which_one_of(if_source, "source")
    return one_of


def validate_rf_source_type(rf_source, rf_source_types, message):
    one_of, value = betterproto.which_one_of(rf_source, "source")
    if one_of not in rf_source_types:
        raise InvalidIdentityException(f"{message} - {one_of}")


def get_rf_source_name(rf_source: RfSource) -> RFInputRFSource:
    one_of, value = betterproto.which_one_of(rf_source, "source")

    if one_of == "rf_up_conv_output":
        try:
            return RFInputRFSource(value.index)
        except KeyError:
            raise ConnectivityException("up converter index is not valid")
    elif one_of == "external_input":
        external_one_of, external_value = betterproto.which_one_of(value, "input")
        if external_one_of == "rf_in_index":
            return RFInputRFSource.RF_in
        else:
            raise ExternalInputTypeException("external input is not valid")
    elif one_of == "constant_source":
        return RFInputRFSource.Off
    else:
        # "synth_output" not supported
        raise RfSourceTypeException("rf source is not supported")


def get_rf_source_from_synth_panel_output(rf_source: RfSource) -> SynthRfOutput:
    one_of, value = betterproto.which_one_of(rf_source, "source")

    if one_of == "synth_output":
        return value
    else:
        # "rf_up_conv_output" and "constant_source"
        # and "external_input" are not supported
        raise RfSourceTypeException("rf source is not supported")


identity_object_to_module_ref_type_mapping = {
    RfUpConvIdentity: OctaveModule.OCTAVE_MODULE_RF_UPCONVERTER,
    RfDownConvIdentity: OctaveModule.OCTAVE_MODULE_RF_DOWNCONVERTER,
    IfDownConvIdentity: OctaveModule.OCTAVE_MODULE_IF_DOWNCONVERTER,
    SynthIdentity: OctaveModule.OCTAVE_MODULE_SYNTHESIZER,
}



from octave_sdk.grpc.quantummachines.octave.api.v1 import (
    IdentifyResponse,
    RfUpConvIdentity,
    RfUpConvIdentityConnectivity,
    IfSource,
    ExternalIfInput,
    RfSource,
    SynthRfOutput,
    SynthRfOutputOutputPort,
    RfUpConvIdentityParameters,
    RfDownConvIdentity,
    RfDownConvIdentityConnectivity,
    UpConvRfOutput,
    ExternalRfInput,
    IfDownConvIdentity,
    IfDownConvIdentityConnectivity,
    IfDownConvIfSource,
    SynthIdentity,
    SynthIdentityConnectivity,
    SynthIdentityParameters,
    SynthIdentityParametersLowFrequencyFilter,
    SynthIdentityParametersParametrizedFilter,
    PanelIdentity,
    PanelIdentityRfOutput,
    PanelIdentitySynthOutput,
    RfDownConvIfSource,
    RfDownConvIfSourceOutputPort,
    ExploreResponse,
    ExploreResponseModuleId,
    ModuleReference,
)


def default_explore_response():
    return ExploreResponse(
        modules=[
            ExploreResponseModuleId(module=ModuleReference(type=1, index=1), id="0025_F"),
            ExploreResponseModuleId(module=ModuleReference(type=1, index=2), id="0024_F"),
            ExploreResponseModuleId(module=ModuleReference(type=1, index=3), id="01170A01F2226006012"),
            ExploreResponseModuleId(module=ModuleReference(type=1, index=4), id="0027_F"),
            ExploreResponseModuleId(module=ModuleReference(type=1, index=5), id="01170A01F2215017712"),
            ExploreResponseModuleId(module=ModuleReference(type=2, index=1), id="2"),
            ExploreResponseModuleId(module=ModuleReference(type=2, index=2), id="3"),
            ExploreResponseModuleId(module=ModuleReference(type=3, index=1), id="0005_A"),
            ExploreResponseModuleId(module=ModuleReference(type=3, index=2), id="0007_A"),
            ExploreResponseModuleId(module=ModuleReference(type=4, index=3), id="0008_B"),
            ExploreResponseModuleId(module=ModuleReference(type=4, index=4), id="22150030_B"),
            ExploreResponseModuleId(module=ModuleReference(type=4, index=5), id="0010_B"),
            ExploreResponseModuleId(module=ModuleReference(type=4, index=6), id="22150023_B"),
            ExploreResponseModuleId(module=ModuleReference(type=5, index=1)),
        ]
    )


def default_identify_response():
    rf_up_converters = [
        RfUpConvIdentity(
            index=1,
            connectivity=RfUpConvIdentityConnectivity(
                i_input=IfSource(external_if_input=ExternalIfInput(i_index=1)),
                q_input=IfSource(external_if_input=ExternalIfInput(q_index=1)),
                lo_input=RfSource(
                    synth_output=SynthRfOutput(index=3, output_port=SynthRfOutputOutputPort.OUTPUT_PORT_MAIN)
                ),
            ),
            parameters=RfUpConvIdentityParameters(0, 0),
        ),
        RfUpConvIdentity(
            index=2,
            connectivity=RfUpConvIdentityConnectivity(
                i_input=IfSource(external_if_input=ExternalIfInput(i_index=2)),
                q_input=IfSource(external_if_input=ExternalIfInput(q_index=2)),
                lo_input=RfSource(
                    synth_output=SynthRfOutput(
                        index=6,
                        output_port=SynthRfOutputOutputPort.OUTPUT_PORT_SECONDARY,
                    )
                ),
            ),
            parameters=RfUpConvIdentityParameters(0, 0),
        ),
        RfUpConvIdentity(
            index=3,
            connectivity=RfUpConvIdentityConnectivity(
                i_input=IfSource(external_if_input=ExternalIfInput(i_index=3)),
                q_input=IfSource(external_if_input=ExternalIfInput(q_index=3)),
                lo_input=RfSource(
                    synth_output=SynthRfOutput(index=6, output_port=SynthRfOutputOutputPort.OUTPUT_PORT_MAIN)
                ),
            ),
            parameters=RfUpConvIdentityParameters(0, 0),
        ),
        RfUpConvIdentity(
            index=4,
            connectivity=RfUpConvIdentityConnectivity(
                i_input=IfSource(external_if_input=ExternalIfInput(i_index=4)),
                q_input=IfSource(external_if_input=ExternalIfInput(q_index=4)),
                lo_input=RfSource(
                    synth_output=SynthRfOutput(
                        index=5,
                        output_port=SynthRfOutputOutputPort.OUTPUT_PORT_SECONDARY,
                    )
                ),
            ),
            parameters=RfUpConvIdentityParameters(0, 0),
        ),
        RfUpConvIdentity(
            index=5,
            connectivity=RfUpConvIdentityConnectivity(
                i_input=IfSource(external_if_input=ExternalIfInput(i_index=5)),
                q_input=IfSource(external_if_input=ExternalIfInput(q_index=5)),
                lo_input=RfSource(
                    synth_output=SynthRfOutput(index=5, output_port=SynthRfOutputOutputPort.OUTPUT_PORT_MAIN)
                ),
            ),
            parameters=RfUpConvIdentityParameters(0, 0),
        ),
    ]

    rf_down_converters = [
        RfDownConvIdentity(
            index=1,
            connectivity=RfDownConvIdentityConnectivity(
                # debug_rf_input_1=None,
                # debug_rf_input_2=None,
                # debug_rf_input_3=None,
                # debug_rf_input_4=None,
                # debug_rf_input_5=None,
                rf_main_input=RfSource(external_input=ExternalRfInput(rf_in_index=1)),
                lo_input_1=RfSource(
                    synth_output=SynthRfOutput(
                        index=4,
                        output_port=SynthRfOutputOutputPort.OUTPUT_PORT_SECONDARY,
                    )
                ),
                lo_input_2=RfSource(
                    synth_output=SynthRfOutput(
                        index=3,
                        output_port=SynthRfOutputOutputPort.OUTPUT_PORT_SECONDARY,
                    ),
                ),
            ),
        ),
        RfDownConvIdentity(
            index=2,
            connectivity=RfDownConvIdentityConnectivity(
                debug_rf_input_1=RfSource(rf_up_conv_output=UpConvRfOutput(1)),
                debug_rf_input_2=RfSource(rf_up_conv_output=UpConvRfOutput(2)),
                debug_rf_input_3=RfSource(rf_up_conv_output=UpConvRfOutput(3)),
                debug_rf_input_4=RfSource(rf_up_conv_output=UpConvRfOutput(4)),
                debug_rf_input_5=RfSource(rf_up_conv_output=UpConvRfOutput(5)),
                rf_main_input=RfSource(external_input=ExternalRfInput(rf_in_index=2)),
                lo_input_1=RfSource(
                    synth_output=SynthRfOutput(
                        index=4,
                        output_port=SynthRfOutputOutputPort.OUTPUT_PORT_MAIN,
                    )
                ),
                lo_input_2=RfSource(
                    synth_output=SynthRfOutput(
                        index=4,
                        output_port=SynthRfOutputOutputPort.OUTPUT_PORT_SYNTH,
                    ),
                ),
            ),
        ),
    ]

    if_down_converters = [
        IfDownConvIdentity(
            index=1,
            connectivity=IfDownConvIdentityConnectivity(
                channel_1_input=IfSource(
                    rf_downconv_if_source=RfDownConvIfSource(
                        index=1, output_port=RfDownConvIfSourceOutputPort.OUTPUT_PORT_I
                    )
                ),
                channel_1_lo_input=IfSource(external_if_input=ExternalIfInput(if_lo_i_index=1)),
                channel_2_input=IfSource(
                    rf_downconv_if_source=RfDownConvIfSource(
                        index=1, output_port=RfDownConvIfSourceOutputPort.OUTPUT_PORT_Q
                    )
                ),
                channel_2_lo_input=IfSource(external_if_input=ExternalIfInput(if_lo_q_index=1)),
            ),
        ),
        IfDownConvIdentity(
            index=2,
            connectivity=IfDownConvIdentityConnectivity(
                channel_1_input=IfSource(
                    rf_downconv_if_source=RfDownConvIfSource(
                        index=2, output_port=RfDownConvIfSourceOutputPort.OUTPUT_PORT_I
                    )
                ),
                channel_1_lo_input=IfSource(external_if_input=ExternalIfInput(if_lo_i_index=2)),
                channel_2_input=IfSource(
                    rf_downconv_if_source=RfDownConvIfSource(
                        index=2, output_port=RfDownConvIfSourceOutputPort.OUTPUT_PORT_Q
                    )
                ),
                channel_2_lo_input=IfSource(external_if_input=ExternalIfInput(if_lo_q_index=2)),
            ),
        ),
    ]

    synthesizers = [
        SynthIdentity(
            index=3,
            connectivity=SynthIdentityConnectivity(
                main_lo_input=RfSource(external_input=ExternalRfInput(lo_input_index=1)),
                secondary_lo_input=RfSource(external_input=ExternalRfInput(demod_lo_input_index=1)),
            ),
            parameters=SynthIdentityParameters(
                low_frequency_filters=[SynthIdentityParametersLowFrequencyFilter(index=0, filter_1="1", filter_2="2")],
                medium_frequency_filter=SynthIdentityParametersParametrizedFilter(),
                high_frequency_filter=SynthIdentityParametersParametrizedFilter(),
            ),
        ),
        SynthIdentity(
            index=6,
            connectivity=SynthIdentityConnectivity(
                main_lo_input=RfSource(external_input=ExternalRfInput(lo_input_index=3)),
                secondary_lo_input=RfSource(external_input=ExternalRfInput(lo_input_index=2)),
            ),
            parameters=SynthIdentityParameters(
                low_frequency_filters=[SynthIdentityParametersLowFrequencyFilter(index=0, filter_1="1", filter_2="2")],
                medium_frequency_filter=SynthIdentityParametersParametrizedFilter(),
                high_frequency_filter=SynthIdentityParametersParametrizedFilter(),
            ),
        ),
        SynthIdentity(
            index=5,
            connectivity=SynthIdentityConnectivity(
                main_lo_input=RfSource(external_input=ExternalRfInput(lo_input_index=5)),
                secondary_lo_input=RfSource(external_input=ExternalRfInput(lo_input_index=4)),
            ),
            parameters=SynthIdentityParameters(
                low_frequency_filters=[SynthIdentityParametersLowFrequencyFilter(index=0, filter_1="1", filter_2="2")],
                medium_frequency_filter=SynthIdentityParametersParametrizedFilter(),
                high_frequency_filter=SynthIdentityParametersParametrizedFilter(),
            ),
        ),
        SynthIdentity(
            index=4,
            connectivity=SynthIdentityConnectivity(
                # secondary_lo_input=None,
                main_lo_input=RfSource(external_input=ExternalRfInput(demod_lo_input_index=2)),
            ),
            parameters=SynthIdentityParameters(
                low_frequency_filters=[SynthIdentityParametersLowFrequencyFilter(index=0, filter_1="1", filter_2="2")],
                medium_frequency_filter=SynthIdentityParametersParametrizedFilter(),
                high_frequency_filter=SynthIdentityParametersParametrizedFilter(),
            ),
        ),
    ]

    panel_identity = PanelIdentity(
        rf_outputs=[
            PanelIdentityRfOutput(index=1, source=RfSource(rf_up_conv_output=UpConvRfOutput(index=1))),
            PanelIdentityRfOutput(index=2, source=RfSource(rf_up_conv_output=UpConvRfOutput(index=2))),
            PanelIdentityRfOutput(index=3, source=RfSource(rf_up_conv_output=UpConvRfOutput(index=3))),
            PanelIdentityRfOutput(index=4, source=RfSource(rf_up_conv_output=UpConvRfOutput(index=4))),
            PanelIdentityRfOutput(index=5, source=RfSource(rf_up_conv_output=UpConvRfOutput(index=5))),
        ],
        if_output_i=[IfSource(if_downconv_if_source=IfDownConvIfSource(0, 0))],
        if_output_q=[IfSource(if_downconv_if_source=IfDownConvIfSource(0, 1))],
        synth_outputs=[
            PanelIdentitySynthOutput(
                index=1,
                source=RfSource(
                    synth_output=SynthRfOutput(
                        index=3,
                        output_port=SynthRfOutputOutputPort.OUTPUT_PORT_SYNTH,
                    )
                ),
            ),
            PanelIdentitySynthOutput(
                index=2,
                source=RfSource(
                    synth_output=SynthRfOutput(
                        index=6,
                        output_port=SynthRfOutputOutputPort.OUTPUT_PORT_SYNTH,
                    )
                ),
            ),
            PanelIdentitySynthOutput(
                index=3,
                source=RfSource(
                    synth_output=SynthRfOutput(
                        index=5,
                        output_port=SynthRfOutputOutputPort.OUTPUT_PORT_SYNTH,
                    )
                ),
            ),
        ],
    )
    return IdentifyResponse(
        rf_up_converters,
        rf_down_converters,
        if_down_converters,
        synthesizers,
        panel_identity,
    )



import dataclasses
import json
import logging
from threading import Lock
from enum import Enum, auto
from typing import Optional, Dict, Tuple, List

import numpy as np

from octave_sdk.batch import BatchSingleton
from octave_sdk._client_request_builder import ClientRequestBuilder
from octave_sdk.connectivity.connectivity import (
    RFInputRFSource,
    OctaveLOSource,
    RFInputLOSource,
    _LoSourceInfo,
    Connectivity,
    convert_rf_in_enum_to_octave_lo_input,
    OctaveOutput,
    _SynthLOInput,
    convert_octave_lo_input_enum_to_rf_in,
    AnalyzerIndex,
    convert_rf_in_lo_by_rf_out,
    SynthOutputDeviceInfo,
    UpOrDownId,
    ModulesSlotsFromIdentity,
)
from octave_sdk.connectivity.connectivity_util import (
    octave_module_to_module_name_mapping,
    slot_index_to_panel_mapping,
)
from octave_sdk._default_identity import default_identify_response
from octave_sdk._errors import InvalidLoSource
from octave_sdk.grpc.quantummachines.octave.api.v1 import (
    IdentifyResponse,
    SynthRfOutputOutputPort,
    RfDownConvUpdateLoInput,
    SynthUpdateMainSource,
    SynthUpdateMainOutput,
    SynthUpdateSecondaryOutput,
    RfDownConvUpdateRfInput,
    RfUpConvUpdateIfInputSelection,
    RfUpConvUpdateFastSwitchMode,
    SynthUpdateSynthOutput,
    ClockUpdate,
    ClockUpdateMode,
    SynthUpdate,
    SynthUpdateReferenceClock,
    SynthUpdateReferenceSource,
    GetVersionResponse,
    RfUpConvUpdate,
    RfDownConvUpdate,
    OctaveModule,
    SingleUpdate,
    MotherboardUpdate,
    IfDownConvUpdateMode,
    IfDownConvUpdateCoupling,
    IfDownConvUpdate,
    ModuleReference,
    SynthUpdateSynthOutputPower,
)
from octave_sdk._octave_client import OctaveClient
from octave_sdk.conflict_manager import SynthConflicts
from octave_sdk.health_monitor import HealthMonitor
from octave_sdk.octave_module_com import _get_synth_state

DEFAULT_STATE = "default"

logger = logging.getLogger("qm")


class ClockType(Enum):
    """ """

    Internal = auto()
    External = auto()
    Buffered = auto()  # for opt


class ClockFrequency(Enum):
    """ """

    MHZ_10 = auto()
    MHZ_100 = auto()
    MHZ_1000 = auto()


DEFAULT_CLOCK_FREQ = ClockFrequency.MHZ_1000


@dataclasses.dataclass(frozen=True)
class RFOutputLoSourceOption:
    label: OctaveLOSource


@dataclasses.dataclass(frozen=True)
class RFInputLoSourceOption:
    label: RFInputLOSource


class IFMode(Enum):
    """ """

    direct = 1
    envelope = 2
    mixer = 3
    off = 4


class RFOutputMode(Enum):
    """ """

    trig_normal = 1
    trig_inverse = 2
    on = 3
    off = 4
    debug = 5


@dataclasses.dataclass
class ClockInfo:
    clock_type: ClockType
    frequency: Optional[ClockFrequency]


class _OctaveContext:
    def __init__(self, client: OctaveClient, connectivity: Connectivity) -> None:
        super().__init__()
        self.client: OctaveClient = client
        self.connectivity: Connectivity = connectivity


def _turn_off_synth_output(crb, synth_index, synth_output, rf_in_index):
    if synth_output == SynthRfOutputOutputPort.OUTPUT_PORT_MAIN:
        crb.synth[synth_index].main_output = SynthUpdateMainOutput.MAIN_OUTPUT_OFF
    elif synth_output == SynthRfOutputOutputPort.OUTPUT_PORT_SECONDARY:
        crb.synth[synth_index].secondary_output = SynthUpdateSecondaryOutput.SECONDARY_OUTPUT_OFF
    elif synth_output == SynthRfOutputOutputPort.OUTPUT_PORT_SYNTH:
        pass
    else:
        raise Exception(f"could not turn off lo source of rf input {rf_in_index}")


def _set_synth_internal_mode(crb, synth_index, synth_output):
    crb.synth[synth_index].main_source = SynthUpdateMainSource.MAIN_SOURCE_SYNTHESIZER
    crb.synth[synth_index].digital_attn = 0
    crb.synth[synth_index].gain = 0xFFFF
    crb.synth[synth_index].synth_output_power = SynthUpdateSynthOutputPower.SYNTH_OUTPUT_POWER_POS5DB
    if synth_output == SynthRfOutputOutputPort.OUTPUT_PORT_MAIN:
        crb.synth[synth_index].main_output = SynthUpdateMainOutput.MAIN_OUTPUT_MAIN
    elif synth_output == SynthRfOutputOutputPort.OUTPUT_PORT_SECONDARY:
        crb.synth[synth_index].secondary_output = SynthUpdateSecondaryOutput.SECONDARY_OUTPUT_MAIN
    elif synth_output == SynthRfOutputOutputPort.OUTPUT_PORT_SYNTH:
        # no need to change any switches
        pass
    else:
        raise Exception("synth output is unknown")


def _set_synth_external_mode(crb, synth_index, synth_output, synth_lo_input):
    synth = crb.synth[synth_index]
    synth.gain = 0xFFFF
    if synth_output == SynthRfOutputOutputPort.OUTPUT_PORT_MAIN:
        if synth_lo_input == _SynthLOInput.Main:
            synth.main_source = SynthUpdateMainSource.MAIN_SOURCE_EXTERNAL
            synth.main_output = SynthUpdateMainOutput.MAIN_OUTPUT_MAIN
        else:
            raise Exception("can not connect input to output")

    elif synth_output == SynthRfOutputOutputPort.OUTPUT_PORT_SECONDARY:
        if synth_lo_input == _SynthLOInput.Main:
            synth.main_source = SynthUpdateMainSource.MAIN_SOURCE_EXTERNAL
            synth.secondary_output = SynthUpdateMainOutput.MAIN_OUTPUT_MAIN
            synth.secondary_output = SynthUpdateMainOutput.MAIN_OUTPUT_MAIN
        elif synth_lo_input == _SynthLOInput.Secondary:
            synth.secondary_output = SynthUpdateSecondaryOutput.SECONDARY_OUTPUT_AUXILARY
        else:
            raise Exception("synth input is unknown")

    else:
        raise Exception("synth output is unknown")


def _set_if_mode(client, index, mode, channel_id):
    crb = ClientRequestBuilder()
    if channel_id == 1:
        channel = crb.ifconv[index].channel1
    else:
        channel = crb.ifconv[index].channel2

    channel.coupling = IfDownConvUpdateCoupling.COUPLING_DC
    if mode == IFMode.direct:
        channel.mode = IfDownConvUpdateMode.MODE_BYPASS
    elif mode == IFMode.envelope:
        channel.mode = IfDownConvUpdateMode.MODE_POWER_DETECT
    elif mode == IFMode.mixer:
        channel.mode = IfDownConvUpdateMode.MODE_MIXER
    elif mode == IFMode.off:
        channel.mode = IfDownConvUpdateMode.MODE_OFF
    else:
        raise ValueError(f"IF Mode {mode} is not supported. only `IFMode` enum values" f" are supported")
    client.update(crb.get_updates())


def _get_if_mode(client, index, channel):
    response: SingleUpdate = client.aquire_modules(
        [ModuleReference(type=OctaveModule.OCTAVE_MODULE_IF_DOWNCONVERTER, index=index)]
    )
    res: IfDownConvUpdate = response.if_down_conv

    if channel == 1:
        mode = res.channel1.mode
    else:
        mode = res.channel2.mode

    if mode == IfDownConvUpdateMode.MODE_BYPASS:
        return IFMode.direct
    elif mode == IfDownConvUpdateMode.MODE_POWER_DETECT:
        return IFMode.envelope
    elif mode == IfDownConvUpdateMode.MODE_MIXER:
        return IFMode.mixer
    elif mode == IfDownConvUpdateMode.MODE_OFF:
        return IFMode.off
    else:
        raise Exception("can not get if mode")


max_gain_graph = np.array(
    [
        [2, 30.587],
        [3, 29.916],
        [4, 27.76],
        [5, 26.986],
        [6, 27.15],
        [7, 26.736],
        [8, 25.348],
        [9, 24.499],
        [10, 24.091],
        [11, 20.947],
        [12, 20.162],
        [13, 21.46],
        [14, 21.141],
        [15, 20.24],
        [16, 19.013],
        [17, 16.761],
        [18, 15.909],
    ]
)


class RFInput:
    """ """

    def __init__(
        self,
        context: _OctaveContext,
        index: int,
    ) -> None:
        super().__init__()
        self._context = context
        self._index = index
        self._possible_lo_sources_first: Dict[
            Tuple[OctaveLOSource, int], _LoSourceInfo
        ] = context.connectivity.get_lo_source_of_rf_input_first(index)
        self._possible_lo_sources_second: Dict[
            Tuple[OctaveLOSource, int], _LoSourceInfo
        ] = context.connectivity.get_lo_source_of_rf_input_second(index)
        self._possible_lo_sources_names = set(
            [s[0] for s in self._possible_lo_sources_first] + [s[0] for s in self._possible_lo_sources_second]
        )
        self._possible_rf_sources = context.connectivity.get_rf_source_of_rf_input(index)

        self._if_index = self._context.connectivity.get_if_by_down_conv(self._index)

    def _get_current_state(self):
        response = self._context.client.aquire_modules(
            [ModuleReference(type=OctaveModule.OCTAVE_MODULE_RF_DOWNCONVERTER, index=self._index)]
        )
        down_conv_state = response.rf_down_conv
        if isinstance(down_conv_state, RfDownConvUpdate):
            return down_conv_state
        else:
            raise Exception("could not get rf input state")

    def get_rf_source(self) -> RFInputRFSource:
        """ """
        rf_source: RfDownConvUpdateRfInput = self._get_current_state().rf_input
        if rf_source == RfDownConvUpdateRfInput.RF_INPUT_DISCONNECT:
            return RFInputRFSource.Off
        elif rf_source == RfDownConvUpdateRfInput.RF_INPUT_MAIN:
            return RFInputRFSource.RF_in
        elif rf_source == RfDownConvUpdateRfInput.RF_INPUT_DEBUG_1:
            return RFInputRFSource.Loopback_RF_out_1
        elif rf_source == RfDownConvUpdateRfInput.RF_INPUT_DEBUG_2:
            return RFInputRFSource.Loopback_RF_out_2
        elif rf_source == RfDownConvUpdateRfInput.RF_INPUT_DEBUG_3:
            return RFInputRFSource.Loopback_RF_out_3
        elif rf_source == RfDownConvUpdateRfInput.RF_INPUT_DEBUG_4:
            return RFInputRFSource.Loopback_RF_out_4
        elif rf_source == RfDownConvUpdateRfInput.RF_INPUT_DEBUG_5:
            return RFInputRFSource.Loopback_RF_out_5
        else:
            raise ValueError(f"Current RF source of rf input {self._index} is unknown")

    def set_rf_source(self, name: RFInputRFSource):
        """

        :param name:
        """

        if name not in self._possible_rf_sources:
            raise ValueError(f"RF source {name} is invalid for RF input {self._index}")

        crb = ClientRequestBuilder()
        if name == RFInputRFSource.RF_in:
            crb.down[self._index].rf_input = RfDownConvUpdateRfInput.RF_INPUT_MAIN
            crb.down[self._index].enabled = True
        elif name == RFInputRFSource.Loopback_RF_out_1:
            crb.down[self._index].rf_input = RfDownConvUpdateRfInput.RF_INPUT_DEBUG_1
            crb.down[self._index].enabled = True
        elif name == RFInputRFSource.Loopback_RF_out_2:
            crb.down[self._index].rf_input = RfDownConvUpdateRfInput.RF_INPUT_DEBUG_2
            crb.down[self._index].enabled = True
        elif name == RFInputRFSource.Loopback_RF_out_3:
            crb.down[self._index].rf_input = RfDownConvUpdateRfInput.RF_INPUT_DEBUG_3
            crb.down[self._index].enabled = True
        elif name == RFInputRFSource.Loopback_RF_out_4:
            crb.down[self._index].rf_input = RfDownConvUpdateRfInput.RF_INPUT_DEBUG_4
            crb.down[self._index].enabled = True
        elif name == RFInputRFSource.Loopback_RF_out_5:
            crb.down[self._index].rf_input = RfDownConvUpdateRfInput.RF_INPUT_DEBUG_5
            crb.down[self._index].enabled = True
        elif name == RFInputRFSource.Off:
            crb.down[self._index].rf_input = RfDownConvUpdateRfInput.RF_INPUT_DISCONNECT
            crb.down[self._index].enabled = False
        else:
            raise ValueError(f"RF source {name} is not supported. only `InputRfSource` enum values" f" are supported")

        self._context.client.update(crb.get_updates())

    def set_lo_source(self, source_name: RFInputLOSource, ignore_shared_errors: bool = False):
        """
        Set LO for the selected RFinput
        @param source_name: The LO type from RFInputLOSource
        @param ignore_shared_errors: Override shared LO error scenarios and replace with warning instead
        """
        self._set_lo_source_of_rf_in(source_name, select=True, ignore_shared_errors=ignore_shared_errors)

    def set_lo_frequency(self, source_name: RFInputLOSource, frequency: float):
        self._set_lo_source_of_rf_in(source_name, frequency=frequency)

    def get_lo_source(self) -> RFInputLOSource:
        """ """
        rf_in_source, synth_index = self._get_rf_in_lo_input()
        return convert_rf_in_lo_by_rf_out(rf_in_source, synth_index, self._context.connectivity.rf_out_by_synth())

    def get_lo_frequency(self) -> float:
        """ """
        rf_in_source, synth_index = self._get_rf_in_lo_input()
        if rf_in_source == RFInputLOSource.Internal or rf_in_source == RFInputLOSource.Analyzer:
            return _get_synth_state(self._context.client, synth_index).synth_output.frequency
        else:
            raise NotImplementedError()

    def _get_synth_external_secondary_input(self, synth_index):
        possible_sources = [
            u
            for u in self._context.connectivity.get_lo_source_by_synth_output(
                synth_index, SynthRfOutputOutputPort.OUTPUT_PORT_SECONDARY
            ).values()
            if u.module_info.synth_index == synth_index and u.module_info.synth_lo_input == _SynthLOInput.Secondary
        ]

        if len(possible_sources) != 1:
            raise ValueError(f"problem getting synth {synth_index} lo source")

        return (
            convert_octave_lo_input_enum_to_rf_in(possible_sources[0].name, synth_index),
            synth_index,
        )

    def _get_synth_external_main_input(self, synth_index):
        possible_sources = [
            u
            for u in self._context.connectivity.get_lo_source_by_synth_output(
                synth_index, SynthRfOutputOutputPort.OUTPUT_PORT_MAIN
            ).values()
            if u.module_info.synth_index == synth_index and u.module_info.synth_lo_input == _SynthLOInput.Main
        ]
        if len(possible_sources) != 1:
            raise ValueError("problem getting synth lo source")

        return (
            convert_octave_lo_input_enum_to_rf_in(possible_sources[0].name, synth_index),
            synth_index,
        )

    def _get_synth_main_lo(self, synth_state, synth_index):
        if synth_state.main_source == SynthUpdateMainSource.MAIN_SOURCE_EXTERNAL:
            return self._get_synth_external_main_input(synth_index)
        elif synth_state.main_source == SynthUpdateMainSource.MAIN_SOURCE_SYNTHESIZER:
            if synth_index == AnalyzerIndex:
                return RFInputLOSource.Analyzer, synth_index
            else:
                return RFInputLOSource.Internal, synth_index

    def _get_rf_in_lo_input(self) -> (RFInputLOSource, int):
        lo_source = self._get_current_state().lo_input
        # validating what is connected to the synth
        (
            synth_index,
            synth_output,
        ) = self._context.connectivity.get_synth_index_of_rf_in_lo(self._index, lo_source)
        synth_state = _get_synth_state(self._context.client, synth_index)
        if synth_output == SynthRfOutputOutputPort.OUTPUT_PORT_MAIN:
            if synth_state.main_output == SynthUpdateMainOutput.MAIN_OUTPUT_MAIN:
                return self._get_synth_main_lo(synth_state, synth_index)
            else:
                return OctaveLOSource, synth_index
        elif synth_output == SynthRfOutputOutputPort.OUTPUT_PORT_SECONDARY:
            if synth_state.secondary_output == SynthUpdateSecondaryOutput.SECONDARY_OUTPUT_MAIN:
                return self._get_synth_main_lo(synth_state, synth_index)
            elif synth_state.secondary_output == SynthUpdateSecondaryOutput.SECONDARY_OUTPUT_AUXILARY:
                return self._get_synth_external_secondary_input(synth_index)
            else:
                return OctaveLOSource, synth_index
        else:
            raise ValueError("Could not get LO of RF input")

    def _set_lo_source_of_rf_in(
        self,
        name: RFInputLOSource,
        frequency: Optional[float] = None,
        select: Optional[bool] = None,
        ignore_shared_errors: bool = False,
        check_conflict: bool = False,
    ):

        (octave_lo_source, synth_index) = convert_rf_in_enum_to_octave_lo_input(
            name,
            self._context.connectivity.synth_by_rf_out(),
            self._context.connectivity.synth_by_lo_source(),
        )

        if (
            octave_lo_source == OctaveLOSource.Internal or octave_lo_source == OctaveLOSource.Off
        ) and synth_index == -1:
            synth_index = self._get_internal_synth_index()

        if (octave_lo_source, synth_index) not in self._possible_lo_sources_first and (
            octave_lo_source,
            synth_index,
        ) not in self._possible_lo_sources_second:
            possible_values = self._get_possible_lo_sources()

            unique_values = set(possible_values)
            raise InvalidLoSource(
                f"{name} is a not a valid LO source of RF input {self._index}. "
                f"possible values are: {','.join(map(str, unique_values))}"
            )

        if (octave_lo_source, synth_index) in self._possible_lo_sources_first:
            input_index = RfDownConvUpdateLoInput.LO_INPUT_1
            update_input_index = RfDownConvUpdateLoInput.LO_INPUT_1
            source_info = self._possible_lo_sources_first[(octave_lo_source, synth_index)]
        else:
            input_index = RfDownConvUpdateLoInput.LO_INPUT_2
            update_input_index = RfDownConvUpdateLoInput.LO_INPUT_2
            source_info = self._possible_lo_sources_second[(octave_lo_source, synth_index)]

        (
            synth_index,
            synth_output,
        ) = self._context.connectivity.get_synth_index_of_rf_in_lo(self._index, input_index)

        # get other input synth id so we can turn it off
        if input_index == RfDownConvUpdateLoInput.LO_INPUT_1:
            other = RfDownConvUpdateLoInput.LO_INPUT_2
        else:
            other = RfDownConvUpdateLoInput.LO_INPUT_1
        (
            other_synth_index,
            other_synth_output,
        ) = self._context.connectivity.get_synth_index_of_rf_in_lo(self._index, other)

        if select is not None and check_conflict:
            conflict_result = SynthConflicts(
                synth_index,
                synth_output,
                octave_lo_source,
                self._context.connectivity,
                self._context.client,
                request_up_or_down=SynthOutputDeviceInfo(index=self._index, up_or_down=UpOrDownId.DownConv),
                ignore_shared_errors=ignore_shared_errors,
            )
            if conflict_result.is_conflict:
                conflict_result.print_error()

        if octave_lo_source == OctaveLOSource.Internal:
            if frequency is not None:
                self._set_frequency(frequency, synth_index)

            if select is not None:
                crb = ClientRequestBuilder()
                crb.down[self._index].lo_input = update_input_index
                _set_synth_internal_mode(crb, synth_index, synth_output)

                self._context.client.update(crb.get_updates())

                if other_synth_index is not None:
                    _turn_off_synth_output(crb, other_synth_index, other_synth_output, self._index)

        elif octave_lo_source == OctaveLOSource.Off:
            crb = ClientRequestBuilder()
            _turn_off_synth_output(crb, synth_index, synth_output, self._index)

            self._context.client.update(crb.get_updates())

        elif octave_lo_source in self._context.connectivity.loopbacks:
            external_synth_output = self._context.connectivity.loopbacks[octave_lo_source]

            if frequency is not None:
                self._set_frequency(
                    frequency,
                    self._context.connectivity.get_synth_index_from_output_port(external_synth_output),
                )

            if select is not None:
                self._set_external(
                    input_index,
                    other_synth_index,
                    other_synth_output,
                    source_info,
                    synth_index,
                    synth_output,
                )

        else:
            if frequency is not None:
                raise ValueError(f"Frequency can not be set for external LO source {name}")
            if select is not None:
                self._set_external(
                    input_index,
                    other_synth_index,
                    other_synth_output,
                    source_info,
                    synth_index,
                    synth_output,
                )

    def _set_frequency(self, frequency, synth_index):
        crb = ClientRequestBuilder()
        crb.synth[synth_index].synth_output = SynthUpdateSynthOutput(frequency=frequency)
        self._context.client.update(crb.get_updates())

    def _set_external(
        self,
        input_index,
        other_synth_index,
        other_synth_output,
        source_info,
        synth_index,
        synth_output,
    ):
        crb = ClientRequestBuilder()
        crb.down[self._index].lo_input = input_index
        _set_synth_external_mode(
            crb,
            synth_index,
            synth_output,
            source_info.module_info.synth_lo_input,
        )
        self._context.client.update(crb.get_updates())
        if other_synth_index is not None:
            _turn_off_synth_output(crb, other_synth_index, other_synth_output, self._index)

    def _get_internal_synth_index(self):
        synth_index = -1
        _, analyzer_synth = convert_rf_in_enum_to_octave_lo_input(
            RFInputLOSource.Analyzer,
            self._context.connectivity.synth_by_rf_out(),
            self._context.connectivity.synth_by_lo_source(),
        )
        synth_possibilities = [
            index for source, index in self._possible_lo_sources_first.keys() if source == OctaveLOSource.Internal
        ] + [index for source, index in self._possible_lo_sources_second.keys() if source == OctaveLOSource.Internal]
        for pos in synth_possibilities:
            if pos != analyzer_synth:
                return pos

        for pos in synth_possibilities:
            return pos
        return synth_index

    def _get_possible_lo_sources(self):
        possible_values = [
            convert_rf_in_lo_by_rf_out(
                convert_octave_lo_input_enum_to_rf_in(e[0], e[1]),
                e[1],
                self._context.connectivity.rf_out_by_synth(),
            )
            for e in self._possible_lo_sources_first.keys()
        ] + [
            convert_rf_in_lo_by_rf_out(
                convert_octave_lo_input_enum_to_rf_in(e[0], e[1]),
                e[1],
                self._context.connectivity.rf_out_by_synth(),
            )
            for e in self._possible_lo_sources_second.keys()
        ]
        return possible_values

    def get_if_mode_i(self) -> IFMode:
        """ """
        return _get_if_mode(self._context.client, index=self._if_index, channel=1)

    def get_if_mode_q(self) -> IFMode:
        """ """
        return _get_if_mode(self._context.client, index=self._if_index, channel=2)

    def set_if_mode_i(self, mode: IFMode):
        """

        :param mode:
        """
        _set_if_mode(self._context.client, self._if_index, mode, channel_id=1)

    def set_if_mode_q(self, mode: IFMode):
        """

        :param mode:
        """
        _set_if_mode(self._context.client, self._if_index, mode, channel_id=2)

    def get_temperature(self) -> float:
        """

        :return: Celsius
        """
        # we have temperature for all modules, but here the important one is down conv
        return self._context.client.monitor().modules[OctaveModule.OCTAVE_MODULE_RF_DOWNCONVERTER][self._index - 1]

    def get_lo_sources(self) -> List[RFInputLoSourceOption]:
        return list(set([RFInputLoSourceOption(source) for source in self._get_possible_lo_sources()]))


class RFOutput:
    """ """

    def __init__(
        self,
        context: _OctaveContext,
        index: int,
    ) -> None:
        super().__init__()
        self._context = context
        self._index = index
        self._possible_lo_sources: Dict[
            Tuple[OctaveLOSource, int], _LoSourceInfo
        ] = context.connectivity.get_lo_source_of_rf_output(index)
        self._possible_lo_sources_names = set([s[0] for s in self._possible_lo_sources])

    def _get_current_state(self):
        response = self._context.client.aquire_modules(
            [ModuleReference(type=OctaveModule.OCTAVE_MODULE_RF_UPCONVERTER, index=self._index)]
        )
        up_conv_state = response.rf_up_conv
        if isinstance(up_conv_state, RfUpConvUpdate):
            return up_conv_state
        else:
            raise Exception("could not get rf output state")

    def get_gain(self) -> float:
        """ """
        raise NotImplementedError()

    def set_gain(self, gain: float, lo_frequency: Optional[float] = None):
        crb = ClientRequestBuilder()

        if lo_frequency is not None:
            frequency = lo_frequency
        else:
            try:
                frequency = self.get_lo_frequency()
            except Exception as e:
                raise Exception(
                    f"Error, set gain to RFOut[{self._index}]"
                    f" has no frequency defined for the lo source."
                    f" Set the LO freq or implicitly add freq to set_gain"
                ) from e

            if frequency is None:
                raise Exception("there is no frequency defined on the lo source")

        self._set_gain(crb, gain, frequency)
        self._context.client.update(crb.get_updates())

    def _set_gain(self, crb, gain: float, frequency: float):
        """
            if no lo_frequency is given, and lo source is internal, will use the
            internal frequency
        :param gain: dB
        :param gain: lo_frequency
        """

        max_gain = np.interp(
            frequency / 1e9,
            max_gain_graph[:, 0],
            max_gain_graph[:, 1],
        )

        max_gain = max_gain - 2  # backoff variance between modules

        max_value_to_reduce = 45

        if gain > max_gain:
            raise ValueError(f"maximum gain for frequency {frequency} is {max_gain}")
        elif int(max_gain - gain) > max_value_to_reduce:
            raise ValueError(f"minimum gain for frequency {frequency} is {max_gain - max_value_to_reduce}")
        elif max_gain - gain < 10:
            crb.up[self._index].input_attn = RfUpConvUpdateIfInputSelection.IF_INPUT_SELECTION_PASS_THROUGH
            self._set_attenuators(crb, gain, max_gain)
        else:
            crb.up[self._index].input_attn = RfUpConvUpdateIfInputSelection.IF_INPUT_SELECTION_ATTENUATE_10DB
            self._set_attenuators(crb, gain + 10, max_gain)

    def _set_attenuators(self, crb, gain, max_gain):
        value_to_reduce = max_gain - gain
        mixer_attn_value = min(value_to_reduce, 15)
        additional_amp_attn_value = value_to_reduce - mixer_attn_value

        crb.up[self._index].enabled = True
        crb.up[self._index].mixer_output_attn = int(2 * mixer_attn_value)  # 15DB
        crb.up[self._index].power_amp_enabled = True
        crb.up[self._index].power_amp_attn = int(2 * additional_amp_attn_value)  # 30DB

    def get_output(self) -> RFOutputMode:
        """ """
        up_conv_state = self._get_current_state()
        mode = up_conv_state.fast_switch_mode
        enabled = up_conv_state.enabled

        if mode == RfUpConvUpdateFastSwitchMode.FAST_SWITCH_MODE_OFF:
            if enabled:
                return RFOutputMode.debug
            else:
                return RFOutputMode.off
        elif mode == RfUpConvUpdateFastSwitchMode.FAST_SWITCH_MODE_DIRECT:
            return RFOutputMode.trig_normal
        elif mode == RfUpConvUpdateFastSwitchMode.FAST_SWITCH_MODE_ON:
            return RFOutputMode.on
        elif mode == RfUpConvUpdateFastSwitchMode.FAST_SWITCH_MODE_INVERTED:
            return RFOutputMode.trig_inverse
        else:
            raise ValueError(f"Current mode of rf output {self._index} is unknown")

    def _set_default_if_needed(self, crb: ClientRequestBuilder):
        up_conv_state = self._get_current_state()
        # TODO enabled = up_conv_state.enabled and up_conv_state.power_amp_enabled
        enabled = up_conv_state.power_amp_enabled
        if not enabled:
            # define default -20 dB on 8Ghz
            self._set_gain(crb, -20, 8e9)

    def set_output(self, mode: RFOutputMode):
        """

        :param mode:
        """
        crb = ClientRequestBuilder()

        if mode == RFOutputMode.off:
            crb.up[self._index].fast_switch_mode = RfUpConvUpdateFastSwitchMode.FAST_SWITCH_MODE_OFF
            crb.up[self._index].enabled = False
        elif mode == RFOutputMode.debug:
            crb.up[self._index].fast_switch_mode = RfUpConvUpdateFastSwitchMode.FAST_SWITCH_MODE_OFF
            self._set_default_if_needed(crb)
            crb.up[self._index].enabled = True
        elif mode == RFOutputMode.on:
            crb.up[self._index].fast_switch_mode = RfUpConvUpdateFastSwitchMode.FAST_SWITCH_MODE_ON
            self._set_default_if_needed(crb)
            crb.up[self._index].enabled = True
        elif mode == RFOutputMode.trig_normal:
            crb.up[self._index].fast_switch_mode = RfUpConvUpdateFastSwitchMode.FAST_SWITCH_MODE_DIRECT
            self._set_default_if_needed(crb)
            crb.up[self._index].enabled = True
        elif mode == RFOutputMode.trig_inverse:
            crb.up[self._index].fast_switch_mode = RfUpConvUpdateFastSwitchMode.FAST_SWITCH_MODE_INVERTED
            self._set_default_if_needed(crb)
            crb.up[self._index].enabled = True
        else:
            raise ValueError(
                f"RF Output Mode {mode} is not supported. only `RFOutputMode` enum" f" values are supported"
            )

        self._context.client.update(crb.get_updates())

    def set_lo_source(self, source_name: OctaveLOSource, ignore_shared_errors: bool = False):
        """
        Set LO for the selected RFoutput
        @param source_name: The LO type from OctaveLOSource
        @param ignore_shared_errors: Override shared LO error scenarios and replace with warning instead
        """
        self._set_lo_source_of_rf_out(source_name, select=True, ignore_shared_errors=ignore_shared_errors)

    def set_lo_frequency(self, source_name: OctaveLOSource, frequency: float):
        if frequency < 2e9 or frequency > 18e9:
            raise ValueError(f"frequency {frequency} is not supported")
        self._set_lo_source_of_rf_out(source_name, frequency=frequency)

    def get_lo_source(self) -> OctaveLOSource:
        """ """
        octave_lo_source, synth_index = self._get_rf_out_lo_input()
        return octave_lo_source

    def get_lo_frequency(self) -> float:
        """ """
        octave_lo_source, synth_index = self._get_rf_out_lo_input()

        if octave_lo_source == OctaveLOSource.Internal:
            return _get_synth_state(self._context.client, synth_index).synth_output.frequency
        else:
            raise Exception(f"could not get frequency of " f"external LO source - RF input {self._index}")

    def _get_synth_external_secondary_input(self, synth_index) -> (OctaveLOSource, int):
        possible_sources = [
            u
            for u in self._possible_lo_sources.values()
            if u.module_info.synth_index == synth_index and u.module_info.synth_lo_input == _SynthLOInput.Secondary
        ]
        if len(possible_sources) != 1:
            raise ValueError(f"problem getting synth {synth_index} lo source")

        return possible_sources[0].name, synth_index

    def _get_synth_external_main_input(self, synth_index) -> (OctaveLOSource, int):
        possible_sources = [
            u
            for u in self._possible_lo_sources.values()
            if u.module_info.synth_index == synth_index and u.module_info.synth_lo_input == _SynthLOInput.Main
        ]
        if len(possible_sources) != 1:
            raise ValueError("problem getting synth lo source")

        return possible_sources[0].name, synth_index

    def _get_synth_main_lo(self, synth_state, synth_index) -> (OctaveLOSource, int):

        if synth_state.main_source == SynthUpdateMainSource.MAIN_SOURCE_EXTERNAL:
            return self._get_synth_external_main_input(synth_index)
        elif synth_state.main_source == SynthUpdateMainSource.MAIN_SOURCE_SYNTHESIZER:
            return OctaveLOSource.Internal, synth_index

    def _get_rf_out_lo_input(self) -> (OctaveLOSource, int):
        # duplicate helper functions because it uses possible values dict
        # validating whats connected to the synth
        (
            synth_index,
            synth_output,
        ) = self._context.connectivity.get_synth_index_of_rf_out_lo(self._index)
        synth_state = _get_synth_state(self._context.client, synth_index)
        if synth_output == SynthRfOutputOutputPort.OUTPUT_PORT_MAIN:
            if synth_state.main_output == SynthUpdateMainOutput.MAIN_OUTPUT_MAIN:
                return self._get_synth_main_lo(synth_state, synth_index)
            else:
                return OctaveLOSource.Off, synth_index
        elif synth_output == SynthRfOutputOutputPort.OUTPUT_PORT_SECONDARY:
            if synth_state.secondary_output == SynthUpdateSecondaryOutput.SECONDARY_OUTPUT_MAIN:
                return self._get_synth_main_lo(synth_state, synth_index)
            elif synth_state.secondary_output == SynthUpdateSecondaryOutput.SECONDARY_OUTPUT_AUXILARY:
                return self._get_synth_external_secondary_input(synth_index)
            else:
                return OctaveLOSource.Off, synth_index
        else:
            # should not happen
            raise ValueError("Could not get LO input of RF output")

    def _set_lo_source_of_rf_out(
        self,
        name: OctaveLOSource,
        frequency: Optional[float] = None,
        select: Optional[bool] = None,
        ignore_shared_errors: bool = False,
        check_conflict: bool = False,
    ):
        if name not in self._possible_lo_sources_names:
            lo_names = map(str, set([e[0] for e in self._possible_lo_sources.keys()]))
            raise InvalidLoSource(
                f"{name} is a not a valid LO source of RF output {self._index}. "
                f"possible values are: {','.join(lo_names)}"
            )

        (
            synth_index,
            synth_output,
        ) = self._context.connectivity.get_synth_index_of_rf_out_lo(self._index)

        if select is not None and check_conflict:
            conflict_result = SynthConflicts(
                synth_index,
                synth_output,
                name,
                self._context.connectivity,
                self._context.client,
                request_up_or_down=SynthOutputDeviceInfo(index=self._index, up_or_down=UpOrDownId.UpConv),
                ignore_shared_errors=ignore_shared_errors,
            )
            if conflict_result.is_conflict:
                conflict_result.print_error()

        if name == OctaveLOSource.Internal:
            # use the synth_index and synth_output_port to set internal mode
            if frequency is not None:
                self._set_frequency(frequency, synth_index)

            if select is not None:
                crb = ClientRequestBuilder()
                _set_synth_internal_mode(crb, synth_index, synth_output)
                self._context.client.update(crb.get_updates())

        elif name == OctaveLOSource.Off:
            crb = ClientRequestBuilder()
            (
                synth_index,
                synth_output,
            ) = self._context.connectivity.get_synth_index_of_rf_out_lo(self._index)
            _turn_off_synth_output(crb, synth_index, synth_output, self._index)
            self._context.client.update(crb.get_updates())

        elif name in self._context.connectivity.loopbacks:
            external_synth_output = self._context.connectivity.loopbacks[name]

            if frequency is not None:
                self._set_frequency(
                    frequency,
                    self._context.connectivity.get_synth_index_from_output_port(external_synth_output),
                )

            if select is not None:
                self._set_external(name, synth_index, synth_output)
        else:
            if frequency is not None:
                raise ValueError(f"Frequency can not be set for external LO source {name}")

            if select is not None:
                self._set_external(name, synth_index, synth_output)

    def _set_frequency(self, frequency, synth_index):
        crb = ClientRequestBuilder()
        crb.synth[synth_index].synth_output = SynthUpdateSynthOutput(frequency=frequency)
        self._context.client.update(crb.get_updates())

    def _set_external(self, name, synth_index, synth_output):
        source_info = self._possible_lo_sources[(name, synth_index)]
        crb = ClientRequestBuilder()
        _set_synth_external_mode(
            crb,
            synth_index,
            synth_output,
            source_info.module_info.synth_lo_input,
        )
        self._context.client.update(crb.get_updates())

    def get_temperature(self) -> float:
        """

        :return: Celsius
        """
        # we have temperature for all modules, but here the important one is upconv
        return self._context.client.monitor().modules[OctaveModule.OCTAVE_MODULE_RF_UPCONVERTER][self._index - 1]

    def get_iq_channels(self):
        return self._context.connectivity.get_iq_channels_for_rf_output(self._index)

    def get_lo_sources(self) -> List[RFOutputLoSourceOption]:
        return list(set([RFOutputLoSourceOption(source[0]) for source in self._possible_lo_sources.keys()]))


class RFInputs:
    """ """

    def __init__(self, context: _OctaveContext, indices: List[int]) -> None:
        super().__init__()
        self._context = context
        self._indices = indices
        self._rf_inputs = [RFInput(self._context, i) for i in self._indices]

    def __getitem__(self, item: int) -> RFInput:
        if item not in self._indices:
            raise IndexError(f"Index was {item} but valid RF " f"inputs are only {','.join(map(str, self._indices))}")
        return self._rf_inputs[self._indices.index(item)]

    def __iter__(self):
        return self._rf_inputs.__iter__()


class RFOutputs:
    """ """

    def __init__(self, context: _OctaveContext, indices: List[int]) -> None:
        super().__init__()
        self._context = context
        self._indices = indices
        self._rf_outputs = [RFOutput(self._context, i) for i in self._indices]

    def __getitem__(self, item) -> RFOutput:
        if item not in self._indices:
            raise IndexError(f"Index was {item} but valid RF " f"outputs are only {','.join(map(str, self._indices))}")
        return self._rf_outputs[self._indices.index(item)]

    def __iter__(self):
        return self._rf_outputs.__iter__()


class Octave:
    """ """

    def __init__(
        self,
        *,
        host: str,
        port: int,
        octave_name: str = None,
        port_mapping: Optional[Dict[OctaveLOSource, OctaveOutput]] = None,
        octave_client: Optional[OctaveClient] = None,
        fan=None,
        identity=None,
        **kwargs,
    ) -> None:
        """

        :param host:
        :param port:
        :param port_mapping: external loop backs, by the label on the box.
        """

        self._host = host
        self._port = port
        self._client = octave_client or OctaveClient(host, port, octave_name)

        self._connectivity_mutex = Lock()  # helps with debugging the health check code
        self._connectivity = None

        # temporary fix:
        self._client.debug_set(True, 1, 20, True, False, True, False, 45, 65, 75)
        if fan is not None:
            self._client.debug_set(monitor_update_fan=False)
            self._set_fan(fan)

        if identity is not None:
            self._identify_response = identity
        else:
            response = self._client.identify()
            if response is not None and len(response.rf_up_converters) > 0:
                self._identify_response: IdentifyResponse = response
            else:
                self._identify_response: IdentifyResponse = default_identify_response()

        self._port_mapping = port_mapping

        # Health check must be called for _reset_connectivity! and should be placed at the very end
        self._health = HealthMonitor(
            self._client,
            octave_name,
            self._reset_connectivity,
            ModulesSlotsFromIdentity(self._identify_response),
            interval_seconds=5,
        )

    def _reset_connectivity(self, explore_result):
        with self._connectivity_mutex:
            self._connectivity: Connectivity = Connectivity(self._identify_response, self._port_mapping, explore_result)

            self._handle_missing_modules()

            self._context = _OctaveContext(self._client, self._connectivity)

            self._rf_inputs = RFInputs(self._context, self._connectivity.rf_inputs_indices())
            self._rf_outputs = RFOutputs(self._context, self._connectivity.rf_outputs_indices())

    def _set_fan(self, fan):
        self._client.update([SingleUpdate(motherboard=MotherboardUpdate(fan))])

    def _get_fan(self):
        response: SingleUpdate = self._context.client.aquire_modules(
            [ModuleReference(type=OctaveModule.OCTAVE_MODULE_MOTHERBOARD, index=1)]
        )
        state = response.motherboard
        return state.fan_speed

    def _handle_missing_modules(self):
        for missing_module in self._connectivity.missing_modules:
            logger.warning(
                f'Octave "{self._client._octave_name}" {octave_module_to_module_name_mapping[missing_module.type]} index {slot_index_to_panel_mapping(missing_module.index, missing_module.type)} is malfunctioning!'
            )

    @staticmethod
    def start_batch_mode():
        BatchSingleton().start_batch_mode()

    @staticmethod
    def end_batch_mode():
        BatchSingleton().end_batch_mode()

    @property
    def rf_inputs(self) -> RFInputs:
        """

        :return:
        """
        return self._rf_inputs

    @property
    def rf_outputs(self) -> RFOutputs:
        """

        :return:
        """
        return self._rf_outputs

    def save_default_state(self, only_clock=False):
        modules = (
            None
            if not only_clock
            else [
                ModuleReference(OctaveModule.OCTAVE_MODULE_SOM, 1),
            ]
            + [ModuleReference(OctaveModule.OCTAVE_MODULE_SYNTHESIZER, i) for i in self._connectivity.synth_indices()]
        )
        result = self._client.save_modules(DEFAULT_STATE, modules, overwrite=True)
        if not result:
            raise Exception("Could not save state")

    def restore_default_state(self):
        self._client.recall(DEFAULT_STATE)

    def snapshot_state(self, name: str, modules: Optional[List] = None):
        if name == DEFAULT_STATE:
            raise ValueError(f"{DEFAULT_STATE} can not be the name of a snapshot")
        result = self._client.save_modules(name, modules, overwrite=True)
        if not result:
            raise Exception("Could not save state")

    def restore_state(self, snapshot_name: str):
        if snapshot_name == DEFAULT_STATE:
            raise ValueError(f"{DEFAULT_STATE} can not be the name of a snapshot")
        self._client.recall(snapshot_name)

    def set_clock(
        self,
        clock_type: ClockType,
        frequency: Optional[ClockFrequency],
        synth_clock=125e6,
    ):
        """
        @param clock_type: Use external or internal or buffered mode
        @param frequency: Frequency of the system clock, not relevant for internal mode
        @param synth_clock: The clock frequency of the synth
        """

        # Not supporting MODE_INTERNAL_USE_1G/MODE_EXT_USE_1G yet
        clock_update = ClockUpdate(synthesizers_clock=synth_clock)
        if clock_type == ClockType.Buffered or (
            clock_type == ClockType.External and frequency == ClockFrequency.MHZ_1000
        ):
            clock_update.mode = ClockUpdateMode.MODE_BUFFERED
            if frequency is None:
                frequency = DEFAULT_CLOCK_FREQ
            elif frequency != DEFAULT_CLOCK_FREQ:
                raise ValueError(
                    "Error, buffered clock frequency must be 1000 MHz, please use External for",
                    ClockFrequency.MHZ_1000.name,
                )
        elif clock_type == ClockType.External:
            clock_update.mode = ClockUpdateMode.MODE_EXTERNAL
            if frequency is None:
                raise ValueError("Error, external clock frequency must be selected")
        elif clock_type == ClockType.Internal:
            clock_update.mode = ClockUpdateMode.MODE_INTERNAL
            if frequency is not None:
                print(
                    "Warning, ignoring",
                    frequency.name,
                    "clock setting and using internal 10MHz clock",
                )
            frequency = ClockFrequency.MHZ_10

        if frequency == ClockFrequency.MHZ_10:
            clock_update.clock_frequency = 10e6
        elif frequency == ClockFrequency.MHZ_100:
            clock_update.clock_frequency = 100e6
        elif frequency == ClockFrequency.MHZ_1000:
            clock_update.clock_frequency = 1000e6

        # TODO if not between 100-150 - should add divider
        synths = [
            SingleUpdate(
                synth=SynthUpdate(
                    index=i,
                    reference_clock=SynthUpdateReferenceClock(
                        source=SynthUpdateReferenceSource.REFERENCE_SOURCE_EXTERNAL,
                        frequency=synth_clock,
                    ),
                )
            )
            for i in self._connectivity.synth_indices()
        ]

        self._client.update([SingleUpdate(clock=clock_update)] + synths)

    def get_clock(self) -> ClockInfo:
        response = self._client.aquire_modules([ModuleReference(type=OctaveModule.OCTAVE_MODULE_SOM, index=1)]).clock
        if isinstance(response, ClockUpdate):
            if response.mode == ClockUpdateMode.MODE_INTERNAL_USE_1G or response.mode == ClockUpdateMode.MODE_INTERNAL:
                clock_type = ClockType.Internal
            elif response.mode == ClockUpdateMode.MODE_EXTERNAL:
                clock_type = ClockType.External
            elif response.mode == ClockUpdateMode.MODE_BUFFERED:
                clock_type = ClockType.Buffered
            else:
                raise ValueError("can not validate clock mode")

            if response.clock_frequency == 10e6:
                frequency = ClockFrequency.MHZ_10
            elif response.clock_frequency == 100e6:
                frequency = ClockFrequency.MHZ_100
            elif response.clock_frequency == 1000e6:
                frequency = ClockFrequency.MHZ_1000
            else:
                raise ValueError("can not validate clock frequency")

            return ClockInfo(clock_type, frequency)

    def get_version(self):
        ver: GetVersionResponse = self._client.version()
        return json.loads(ver.version)["version"]

    def reset(self):
        """
        Will reset the entire Octave HW to default off state
        Warning, will block the code until reset completes
        """
        if self._client.reset():
            self._health.run_once()  # Reset the connectivity in case something changed after reset
            return True

        return False



import asyncio
import logging
import threading
from typing import Callable, Optional

import grpclib.exceptions
from grpclib.exceptions import StreamTerminatedError

from octave_sdk.octave_async_thread import EventLoopThread
from octave_sdk.grpc.quantummachines.octave.api.v1 import HealthRequest, GetVersionRequest

from octave_sdk._octave_client import OctaveClient, ExploreResult, MonitorResult

logger = logging.getLogger("qm")


NUM_CONNECTION_RETRIES = 3


class HealthClient:
    # Creates a constant gRPC stream to the Server, will call the "health_update_callback" on new health result
    # health_update_callback args should be (ExploreResult, MonitorResult)
    def __init__(
        self,
        interval: int,
        client: OctaveClient,
        health_update_callback: Callable[[ExploreResult, MonitorResult], None],
    ):
        self._interval = interval
        self._client = client
        self._callback = health_update_callback
        self._stream_running = False
        self._grace_stop = False
        self._connection_lost = 0

        self._future: Optional[asyncio.Task] = None

        self._stop_event = threading.Event()
        self._event_loop = EventLoopThread().loop

    def start(self):
        self._future = asyncio.run_coroutine_threadsafe(self._health_loop(), loop=self._event_loop)

    def stop(self):
        self._grace_stop = True
        self._stop_event.set()
        if self._future:
            self._future.cancel()

    async def _health_loop(self):
        while not self._stop_event.is_set():
            try:
                await self._health_monitor()
            except (asyncio.CancelledError, StreamTerminatedError, ConnectionRefusedError):
                pass
            except grpclib.exceptions.GRPCError as e:
                if e.status == grpclib.const.Status.UNIMPLEMENTED:
                    # In case of health monitor not implemented, we do not want to print messages of disconnection
                    logger.warning(f'Octave "{self._client.octave_name}" does not support live monitoring')
                    self._stream_running = False
                    return
            finally:
                if not self._grace_stop and self._stream_running:
                    self._connection_lost += 1
                    logger.error(f'Octave "{self._client.octave_name}" lost monitor connection')
                    if self._connection_lost == NUM_CONNECTION_RETRIES:
                        logger.error(f"Failed {NUM_CONNECTION_RETRIES} times to connect Octave aborting")

    async def _health_monitor(self):
        self._stream_running = False
        try:
            await self._client._service.get_version(GetVersionRequest())
            self._health_answers = self._client._service.health(
                HealthRequest(monitor_interval_seconds=self._interval, stop_stream=False)
            )
            self._stream_running = True
            if self._connection_lost:
                self._connection_lost = 0
                logger.info(f'Octave "{self._client.octave_name}" restored monitor connection')

            # FYI infinite loop here until server stream ends or task cancel
            async for response in self._health_answers:
                self._callback(ExploreResult(response.explore), MonitorResult(response.monitor))
        except (asyncio.CancelledError, StreamTerminatedError, ConnectionRefusedError):
            pass

    def run_once(self):
        explore_result = self._client.explore()
        monitor_result = self._client.monitor()
        self._callback(explore_result, monitor_result)



import abc
from collections import defaultdict
from typing import Callable, Dict, Hashable, Tuple

from octave_sdk.grpc.quantummachines.octave.api.v1 import SingleUpdate
from octave_sdk.singleton import Singleton


class BatchSingleton(metaclass=Singleton):
    def __init__(self):
        self._batch_mode = False
        self._start_batch_callbacks: Dict[Hashable, Callable[[], None]] = {}
        self._end_batch_callbacks: Dict[Hashable, Callable[[], None]] = {}

        self._cached_updates: Dict[int, Dict[Tuple[int, str], SingleUpdate]] = defaultdict(dict)
        self._cached_modules: Dict[int, Dict[Tuple[int, str], SingleUpdate]] = defaultdict(dict)

    @property
    def is_batch_mode(self):
        return self._batch_mode

    def start_batch_mode(self):
        if not self._batch_mode:
            for callback in self._start_batch_callbacks.values():
                callback()
        self._batch_mode = True

    def end_batch_mode(self):
        if self._batch_mode:
            for callback in self._end_batch_callbacks.values():
                callback()
            self._cached_updates = defaultdict(dict)
            self._cached_modules = defaultdict(dict)
        self._batch_mode = False

    def set_cached_modules(self, obj: Hashable, modules: Dict[Tuple[int, str], SingleUpdate]):
        self._cached_modules[hash(obj)] = modules

    def set_cached_updates(self, obj: Hashable, modules: Dict[Tuple[int, str], SingleUpdate]):
        self._cached_updates[hash(obj)] = modules

    def get_cached_updates(self, obj: Hashable) -> Dict[Tuple[int, str], SingleUpdate]:
        return self._cached_updates[hash(obj)]

    def get_cached_modules(self, obj: Hashable) -> Dict[Tuple[int, str], SingleUpdate]:
        return self._cached_modules[hash(obj)]

    def register_start_batch_callback(self, obj: Hashable, callback: Callable[[], None]):
        self._start_batch_callbacks[hash(obj)] = callback

    def register_end_batch_callback(self, obj: Hashable, callback: Callable[[], None]):
        self._end_batch_callbacks[hash(obj)] = callback

    def unregister_start_batch_callback(self, obj: Hashable):
        self._start_batch_callbacks.pop(hash(obj))

    def unregister_end_batch_callback(self, obj: Hashable):
        self._end_batch_callbacks.pop(hash(obj))


class Batched(metaclass=abc.ABCMeta):
    def __init__(self):
        BatchSingleton().register_start_batch_callback(self, self._start_batch_callback)
        BatchSingleton().register_end_batch_callback(self, self._end_batch_callback)

    def __del__(self):
        BatchSingleton().unregister_start_batch_callback(self)
        BatchSingleton().unregister_end_batch_callback(self)

    def _start_batch_callback(self):
        pass

    def _end_batch_callback(self):
        pass

    @abc.abstractmethod
    def __hash__(self):
        pass
